'models.py for django-donations'

from __future__ import unicode_literals
from django.db import models
from django.db.utils import OperationalError, ProgrammingError
from django.contrib.auth import get_user_model
from django.utils import timezone
from djmoney.models.fields import MoneyField

from timedelta.fields import TimedeltaField

from importlib import import_module
import logging

from donations import app_settings

logger = logging.getLogger(__name__)


class Frequency(models.Model):
    """(Frequency description)"""
    name = models.CharField(max_length=100)
    # this should be a duration field native in 1.8
    # https://pypi.python.org/pypi/django-timedeltafield
    interval = TimedeltaField()  # this should be celery compatible - should allow for one off vs repeat

    class Meta:  # pylint: disable=C1001
        verbose_name_plural = 'Frequencies'

    def __unicode__(self):
        return "{} ({})".format(self.name, self.interval)


# based on settings auto create db entries? so user does not need to repeat entering info?
class DonationProvider(models.Model):
    """(DonationProvider description)"""

    name = models.CharField(max_length=100, default='BlankProvider')
    description = models.CharField(max_length=255, default='')
    klass = models.CharField(max_length=255, default='')

    def get_provider_class(self):
        """get the class of the donation provider.
        I have hardcoded the first bit to prevent any module being imported"""
        module_name, klass_name = self.klass.rsplit('.', 1)
        module = import_module('donations.providers.{}'.format(module_name))
        return getattr(module, klass_name)

    def __unicode__(self):
        return self.name

DONATION_SOURCE = ["DirectDonations", "SponsorshipDonations", "Ipdd", "Sms"]
DONATION_SOURCE = [(i, i) for i in DONATION_SOURCE]

DONATION_STATUSES = ["Accepted", "Rejected", "Cancelled", "Refunded", "Pending", "Unverified"]
DONATION_STATUSES = [(i, i) for i in DONATION_STATUSES]


class Donation(models.Model):
    """(Abstract representation of a Donation)"""

    # https://github.com/django-money/django-money
    amount = MoneyField(max_digits=10, decimal_places=2, default_currency='GBP')
    provider = models.ForeignKey(DonationProvider, related_name='donations')
    frequency = models.ForeignKey(Frequency, related_name='donations')
    datetime = models.DateTimeField(default=timezone.now)
    donor = models.ForeignKey(get_user_model(), blank=True, null=True, related_name='donations')
    donor_display_name = models.CharField(blank=True, null=True, max_length=255,
                                          help_text="display name returned by the provider")
    status = models.CharField(default='Unverified', choices=DONATION_STATUSES, max_length=50)
    is_verified = models.BooleanField(default=False)
    finished_uri = models.URLField(blank=True)

    provider_ref = models.CharField(blank=True, max_length=100)
    local_amount = MoneyField(max_digits=10, decimal_places=2, default_currency='GBP')
    message = models.CharField(blank=True, null=True, max_length=255)
    est_tax_reclaim = models.DecimalField(blank=True, null=True, max_digits=10, decimal_places=2)
    provider_source = models.CharField(blank=True, null=True, max_length=50, choices=DONATION_SOURCE,
                                       help_text="source of the donation from within a provider")

    def __unicode__(self):
        if self.local_amount:
            return "{} at {}".format(self.local_amount, self.datetime.strftime('%Y/%m/%d %H:%M:%S'))
        return "{} at {}".format(self.amount, self.datetime.strftime('%Y/%m/%d %H:%M:%S'))

    def get_provider(self):
        return self.provider.get_provider_class()(self)

    def get_value(self):
        return self.amount.amount

    def get_currency(self):
        return self.amount.currency

    def donate(self, verify_uri):
        return self.get_provider().donate(verify_uri)

    def verify_donation(self, request):
        self.is_verified = self.get_provider().verify(request)
        self.save()

    def set_money_field(self, field, tuple_content):
        for content in tuple_content:
            if not content:
                logger.warning("Not setting field %s: incomplete data provided %s", field, tuple_content)
                return
        setattr(self, field, tuple_content)

    def set_amount(self, value, currency):
        self.set_money_field('amount', (value, currency))

    def set_local_amount(self, value, currency):
        self.set_money_field('local_amount', (value, currency))


def load_frequencies():
    frequencies = getattr(app_settings, 'DONATION_FREQUENCIES', {})
    logger.info('Loading Donation Frequencies')
    for name, period in frequencies.items():
        try:
            dp, created = Frequency.objects.get_or_create(name=name, interval=period)
            if created:
                logger.info('Loaded %s with interval of %s', name, period)
            else:
                logger.info('Frequency %s with interval of %s already exists', name, period)
        except (OperationalError, ProgrammingError) as exc:
            logger.warning("Could not load the Frequency model instance due to %s", exc)


def load_providers():
    providers = getattr(app_settings, 'DONATION_PROVIDERS', {})
    logger.info('Loading Donation Providers')
    for name, klass in providers.items():
        try:
            dp, created = DonationProvider.objects.get_or_create(name=name, klass=klass)
            if created:
                logger.info('Loaded %s called %s', klass, name)
            else:
                logger.info('Provider %s called %s already exists', klass, name)
        except (OperationalError, ProgrammingError) as exc:
            logger.warning("Could not load the DonationProvider model instance due to %s", exc)


load_providers()
load_frequencies()



'Mixins to make creating donations easy - here is the restriction on choices for amounts, not in the model'


from django.forms import ModelForm, URLField, HiddenInput, Select

from djmoney.forms.widgets import MoneyWidget
from djmoney.forms.fields import MoneyField

from donations.models import Donation


class DonationForm(ModelForm):
    finished_uri = URLField(widget=HiddenInput())

    class Meta:  # pylint: disable=C1001
        model = Donation
        fields = ('amount', 'provider', 'frequency', 'finished_uri')

    def __init__(self, *args, **kwargs):
        super(DonationForm, self).__init__(*args, **kwargs)
        if hasattr(self, 'amounts'):
            self.amounts = [(i, i) if type(i) in [int, float] else i for i in self.amounts]
            # currency_field = ChoiceField(choices=CURRENCY_CHOICES)
            donate_widget = MoneyWidget(amount_widget=Select(choices=self.amounts))
            self.fields['amount'] = MoneyField(currency_widget=donate_widget)



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Frequency'
        db.create_table('donations_frequency', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('interval', self.gf('timedelta.fields.TimedeltaField')()),
        ))
        db.send_create_signal('donations', ['Frequency'])

        # Adding model 'DonationProvider'
        db.create_table('donations_donationprovider', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(default='BlankProvider', max_length=100)),
            ('description', self.gf('django.db.models.fields.CharField')(default='', max_length=255)),
            ('klass', self.gf('django.db.models.fields.CharField')(default='', max_length=255)),
        ))
        db.send_create_signal('donations', ['DonationProvider'])

        # Adding model 'Donation'
        db.create_table('donations_donation', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('amount_currency', self.gf('djmoney.models.fields.CurrencyField')(default='GBP')),
            ('amount', self.gf('djmoney.models.fields.MoneyField')(max_digits=10, decimal_places=2, default_currency='GBP')),
            ('provider', self.gf('django.db.models.fields.related.ForeignKey')(related_name='donations', to=orm['donations.DonationProvider'])),
            ('frequency', self.gf('django.db.models.fields.related.ForeignKey')(related_name='donations', to=orm['donations.Frequency'])),
            ('datetime', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
            ('donor', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='donations', null=True, to=orm['auth.User'])),
            ('donor_display_name', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('status', self.gf('django.db.models.fields.CharField')(default='Unverified', max_length=50)),
            ('is_verified', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('finished_uri', self.gf('django.db.models.fields.URLField')(max_length=200, blank=True)),
            ('provider_ref', self.gf('django.db.models.fields.CharField')(max_length=100, blank=True)),
            ('local_amount_currency', self.gf('djmoney.models.fields.CurrencyField')(default='GBP')),
            ('local_amount', self.gf('djmoney.models.fields.MoneyField')(max_digits=10, decimal_places=2, default_currency='GBP')),
            ('message', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('est_tax_reclaim', self.gf('django.db.models.fields.DecimalField')(null=True, max_digits=10, decimal_places=2, blank=True)),
            ('provider_source', self.gf('django.db.models.fields.CharField')(max_length=50, null=True, blank=True)),
        ))
        db.send_create_signal('donations', ['Donation'])


    def backwards(self, orm):
        # Deleting model 'Frequency'
        db.delete_table('donations_frequency')

        # Deleting model 'DonationProvider'
        db.delete_table('donations_donationprovider')

        # Deleting model 'Donation'
        db.delete_table('donations_donation')


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'related_name': "'user_set'", 'blank': 'True', 'to': "orm['auth.Group']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'related_name': "'user_set'", 'blank': 'True', 'to': "orm['auth.Permission']"}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'donations.donation': {
            'Meta': {'object_name': 'Donation'},
            'amount': ('djmoney.models.fields.MoneyField', [], {'max_digits': '10', 'decimal_places': '2', 'default_currency': "'GBP'"}),
            'amount_currency': ('djmoney.models.fields.CurrencyField', [], {'default': "'GBP'"}),
            'datetime': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'donor': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'donations'", 'null': 'True', 'to': "orm['auth.User']"}),
            'donor_display_name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'est_tax_reclaim': ('django.db.models.fields.DecimalField', [], {'null': 'True', 'max_digits': '10', 'decimal_places': '2', 'blank': 'True'}),
            'finished_uri': ('django.db.models.fields.URLField', [], {'max_length': '200', 'blank': 'True'}),
            'frequency': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'donations'", 'to': "orm['donations.Frequency']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_verified': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'local_amount': ('djmoney.models.fields.MoneyField', [], {'max_digits': '10', 'decimal_places': '2', 'default_currency': "'GBP'"}),
            'local_amount_currency': ('djmoney.models.fields.CurrencyField', [], {'default': "'GBP'"}),
            'message': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'provider': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'donations'", 'to': "orm['donations.DonationProvider']"}),
            'provider_ref': ('django.db.models.fields.CharField', [], {'max_length': '100', 'blank': 'True'}),
            'provider_source': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True', 'blank': 'True'}),
            'status': ('django.db.models.fields.CharField', [], {'default': "'Unverified'", 'max_length': '50'})
        },
        'donations.donationprovider': {
            'Meta': {'object_name': 'DonationProvider'},
            'description': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '255'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'klass': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '255'}),
            'name': ('django.db.models.fields.CharField', [], {'default': "'BlankProvider'", 'max_length': '100'})
        },
        'donations.frequency': {
            'Meta': {'object_name': 'Frequency'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'interval': ('timedelta.fields.TimedeltaField', [], {}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        }
    }

    complete_apps = ['donations']



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf.urls import include, url, patterns
from .views import DonateAPI, VerifyAPI

api_urls = patterns(
    '',
    url(r'^donate/$', DonateAPI.as_view(), name="donate"),
    url(r'^verify/(?P<pk>[0-9]+)$', VerifyAPI.as_view(), name="verify"),
)

view_urls = patterns(
    '',
    # url(r'^donate/$', DonateView.as_view(), name="donate"),
)

donations = patterns(
    '',
    url(r'^api/', include(api_urls, namespace="api")),
    url(r'^', include(view_urls)),
)

urlpatterns = patterns(
    '',
    url(r'^', include(donations, namespace="donations"))
)



from django.contrib import admin
from donations.models import Donation, Frequency, DonationProvider


class DonationAdmin(admin.ModelAdmin):
    list_display = ('amount', 'provider', 'frequency', 'datetime', 'donor', 'is_verified')


class FrequencyAdmin(admin.ModelAdmin):
    list_display = ('name', 'interval')


class DonationProviderAdmin(admin.ModelAdmin):
    list_display = ('name', 'description', 'klass')


admin.site.register(Donation, DonationAdmin)
admin.site.register(Frequency, FrequencyAdmin)
admin.site.register(DonationProvider, DonationProviderAdmin)



'Just Giving Provider for Donations'

from __future__ import unicode_literals
from urllib import quote_plus, urlencode
import logging
import requests

from donations.providers.base import DonationProvider
from donations import app_settings

logger = logging.getLogger(__name__)


class SimpleDonationProvider(DonationProvider):

    charity_id = app_settings.JUST_GIVING_CHARITY_ID
    web_url = app_settings.JUST_GIVING_WEB_URL
    api_url = app_settings.JUST_GIVING_API_URL
    app_id = app_settings.JUST_GIVING_APP_ID

    def get_exit_url(self, verify_uri):
        if self.donation_status == 'Unverified':
            return '{}?donation_id=JUSTGIVING-DONATION-ID'.format(quote_plus(verify_uri))
        return quote_plus(self.donation.redirect_uri)

    def donate(self, verify_uri):
        query_params = {
            'currency': self.get_currency(),
            'amount': self.get_value(),
            'exitUrl': self.get_exit_url(verify_uri),
            'reference': self.donation_reference()
        }
        uri = '{}/4w350m3/donation/direct/charity/{}?{}'.format(self.web_url, self.charity_id, urlencode(query_params))
        return uri

    def donation_reference(self):
        return unicode(self.donation.id)

    def verify(self, request):
        donation_id = request.GET.get('donation_id', '')
        url = '{host}/{appId}/v1/donation/{donationId}'.format(host=self.api_url,
                                                               appId=self.app_id,
                                                               donationId=donation_id)
        resp = requests.get(url, headers={"Content-Type": "application/json"})
        data = resp.json()
        logger.info("JustGiving - Verifying with data = %s", data)
        if data['thirdPartyReference'] == self.donation_reference():
            self.donation.status = data['status']
            if data['status'] == "Accepted":
                # assert on the amount ?
                assert float(self.donation.amount.amount) == float(data['donorLocalAmount'])
                self.donation.message = data['message']
                self.donation.est_tax_reclaim = data.get('estimatedTaxReclaim', 0)
                self.donation.provider_source = data.get('source', None)
                self.donation.set_local_amount(data.get('donorLocalAmount', None),
                                               data.get('donorLocalCurrencyCode', None))
                self.donation.set_amount(data.get('amount', None),
                                         data.get('currencyCode', None))
                self.donation.provider_ref = data.get('donationRef', None)
                self.donation.donor_display_name = data.get('donorDisplayName', None)
                self.donation.save()
                return True
            self.donation.save()
        return False

# {
#     "amount": "2.00",
#     "currencyCode": "GBP",
#     "donationDate": "\/Date(1444216076464+0000)\/",
#     "donationRef": null,
#     "donorDisplayName": "Awesome Guy",
#     "donorLocalAmount": "2.75",
#     "donorLocalCurrencyCode": "EUR",
#     "donorRealName": "Peter Queue",
#     "estimatedTaxReclaim": 0.56,
#     "id": 1234,
#     "image": "",
#     "message": "Hope you like my donation. Rock on!",
#     "source": "SponsorshipDonations",
#     "status": "Accepted",
#     "thirdPartyReference": "1234-my-sdi-ref"
# }



from __future__ import unicode_literals
from urlparse import urlsplit, urlunsplit, parse_qs
from urllib import urlencode

from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
from django.views.generic import CreateView

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status

from donations.models import Donation
from donations.serializers import DonationSerializer


class DonateAPI(APIView):

    def post(self, request):
        # take the required parameters and create the model
        serializer = DonationSerializer(data=request.data)
        if serializer.is_valid():
            ser = serializer.save()
            if self.request.user.is_authenticated():
                ser.donor = self.request.user
                ser.save()
            verify_uri = request.build_absolute_uri(reverse('donations:api:verify', kwargs={'pk': ser.id}))
            redirect_uri = ser.donate(verify_uri)
            return HttpResponseRedirect(redirect_uri)
            # return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def get(self, request):
        donations = Donation.objects.all()
        serializer = DonationSerializer(donations, many=True)
        return Response(serializer.data)


class DonateView(CreateView):
    model = Donation

    def get_form_class(self):
        form_class = super(DonateView, self).get_form_class()
        if hasattr(self, 'amounts'):
            form_class.amounts = self.amounts
        return form_class

    def form_valid(self, form):
        self.object = form.save()
        if self.request.user.is_authenticated():
            self.object.donor = self.request.user
            self.object.save()
        verify_uri = self.request.build_absolute_uri(reverse('donations:api:verify', kwargs={'pk': self.object.id}))
        redirect_uri = self.object.donate(verify_uri)
        return HttpResponseRedirect(redirect_uri)

    def get_context_data(self, **kwargs):
        context = super(DonateView, self).get_context_data(**kwargs)
        context['form'].fields['finished_uri'].initial = self.request.build_absolute_uri(self.set_finished_uri())
        return context

    def set_finished_uri(self):
        'This should be overridden with logic for unique url'
        return '/'


class VerifyAPI(APIView):

    def get(self, request, pk):
        # TODO: url probably needs to be passed through to model/verify method
        donation = Donation.objects.get(pk=pk)
        if donation.status == 'Unverified' and not donation.is_verified:
            # verify it
            donation.verify_donation(request)
        # return redirect + unverfied or verified
        scheme, netloc, path, query_string, frag = urlsplit(donation.finished_uri)
        query_params = parse_qs(query_string)
        query_params['verified'] = [donation.is_verified]
        query_params['ref'] = donation.get_provider().donation_reference()
        url = urlunsplit((scheme, netloc, path, urlencode(query_params, doseq=True), frag))
        return HttpResponseRedirect(url)
